import numpy as np
import tqdm as tq

from .sol_ode import sol_ode
from .delay import delay

def PID_controller(model, var0, t, param, ylimit, w, y0 = 0, PID = [0,0,0], Tt = 0, x_ind = -1, y_ind = -1, n = 1, n_meas = 0):

    """
    Creates a PID controller to adjust the control member y of the input ODE function,
    calculates the controlled variable x using the adjusted parameters.

    Args:
        model (callable(y,t,...)): The function computes the derivative of y at t.
        var0 (array): Initial condition of var.
        t (array): A sequence of time points for which to solve for y.
        param (array, optional): Input parameters to model function.           
        ylimit (array, optional): Lower and upper setpoint for controlled member y (actuator).               
        w (float or int, optional): The setpoint w of the controlled variable x.
        y0 (int, optional): First value of the controlled member y (actuator). Default is 0.
        PID (array, optional): Parameter of the PID controller [KP,KI,KD]. Default is 0.
        Tt (float, optional): Specified dead time.

        x_ind (int, optional): Index of the controlled element x in var0. Default is -1.
        y_ind (int, optional): Index of control member y in param. Default is -1.
        
        n (int, optional): For a cascade of CSTRS, number of total CSTRs n. Default is 1.
        n_meas (int,optional): The reactor in which the controlled variable x measured. Default is the last reactor.

    Returns:
        results: The solved differential equation using the controller.
        y_final: The log of the adjusted control member y.
        e_final: The log of the offset.
    """

#    warnings.filterwarnings("ignore")


    # create the y value list with initial value.
    y = y0
    y_final = np.zeros(len(t))
    x = np.zeros(len(t)-1)
    e_final = np.zeros(len(t))
    e = w
    KP, KI, KD = PID
    Int_e = 0
    
    # empty dataframe for saving results
    results = list(range(len(t)))

    #  store the value calculated from last time.
    for i in tq.tqdm(range(len(t)-1),position=0,leave=True):
        
        y_final[i] = y
        param[y_ind] = y

        #   get the solved results and get the new kla
        df = sol_ode(model, var0, t[i:i+2], param)

        # save the solved results, drop the duplicated value with the same time index,
        # only keeping the last one.
        
        
        if n == 1:
            results[i] = np.transpose(df)[-1]
            var0 = np.transpose(df)[-1]
            x[i] = df[x_ind][-1]
        else:
            results[i] =np.array(df)[:,-1]
            var0 = np.array(df)[:,-1]
            x[i] = df[x_ind][-1,n_meas-1]
            
        
        if Tt != 0:
            x_delay = delay(t, x, Tt, value = 0)

        else:
            x_delay = x

        e_temp = w - x_delay[i]
        # D component
        D = KD * (e - e_temp) / (t[i+1]-t[i])
        # P component
        e = e_temp
        e_final[i] = e_temp
        P = KP * e
        if y<ylimit[1] and y>ylimit[0]:
            Int_e = Int_e + e *(t[i+1]-t[i])
        else: Int_e = Int_e + 0

        # I component
        I = KI*Int_e
        
        
        
        y = y0 + P + I - D
        y = max(min(y, ylimit[1]), ylimit[0])
        
    results[len(t)-1] = results[len(t)-2]
    e_final[len(t)-1] = e_final[len(t)-2]
    y_final[len(t)-1] = y_final[len(t)-2]
    
    if n != 1: 
        r = np.array(results)
        results = r.transpose(1,0,2)
    else: results = np.transpose(results)
    return results, y_final, e_final


import numpy as np
import scipy as sp

def curve_fit(model,data,time,param):

    """
    Fits model to given data with minimization of residuals. 
    
    Args:
        model (callable): Models the data. It must take the independent variable as the first argument and the parameters (as an array) as the second argument.
        data (array): The dependent (measured) data.
        time (array): The independent variable when the data is measured.
        param (array): Parameters of the model
    Returns:
        res.x (array): Optimal values for the parameters so that the sum of the squared residuals between measured and modelled data is minimized.
    """
    
    # Define optimization function
    def fit(param):
        sol = model(time,param)
        return np.linalg.norm(sol[~np.isnan(data)]-data[~np.isnan(data)])

    # Returns initial and parameter values that fit the data perfectly.
    res = sp.optimize.minimize(fit,param)

    return res.x


import numpy as np

def get_index(a, b):

    """
    Returns the index or indices of given elements in list.

    Args:
        a (list or float): Elements to be searched.
        b (list): List for searching.

    Returns:
        inx (list or int): The index or indices of the given elements a found in list b.
        inx_converse (list or int): The index or indices of the given elements a in list a.
    """

    inx = []
    inx_converse = []
    # Check the type of the given element, whether it is a single one or several items.
    if isinstance(a, list) | isinstance(a, np.ndarray):
        for x in a:
            try:
                inx.append(b.index(x))
                inx_converse.append(a.index(x))
            except ValueError:
                pass
        return inx, inx_converse
    else:
        try:
            inx.append(b.index(a))
        except ValueError:
            pass
        return inx


import numpy as np
import tqdm as tq

from .sol_ode import sol_ode
from .delay import delay
from .two_position import two_position

def two_pos_controller(model, var0, t, param, Tt, xlimit, ylimit, x_ind = -1, y_ind = -1, n = 1, n_meas = 0):
    
    """
    Creates a two position controller to adjust the parameters of the input ODE function,
    calculates the controlled variable x with the adjusted parameters.

    Args:
        model (callable(y,t,...)): The function computes the derivative of y at t.
        time (array): A sequence of time points for which to solve for y.
        xlimit (list): Setpoints for controlled variable x.
        ylimit (list): Setpoints for control member y (actuator).
        Tt (float): Specified dead time.
        
        x_ind (int, optional): Index of the controlled element x in var0. Default is -1.
        y_ind (int, optional): Index of control member y in param. Default is -1.
        
        n (int, optional): For a cascade of CSTRS, number of total CSTRs n. Default is 1.
        n_meas (int,optional): The reactor in which the controlled variable x measured. Default is the last reactor.
    
    Returns:
        results: The solved differential equation using the controller.
        y_final: The log of the adjusted control member y.
    """
    
    # create the y value list with initial value.
    y = ylimit[0]
    y_final = np.zeros(len(t))
    x = np.zeros(len(t))

    # empty dataframe for saving results
    results = list(range(len(t)))

    #  store the value calculated from last time.
    for i in tq.tqdm(range(len(t)-1),position=0,leave=True):
        
        y_final[i] = y
        param[y_ind] = y

        #   get the solved results and get the new kla
        df = sol_ode(model, var0, t[i:i+2], param)

        # save the solved results, drop the duplicated value with the same time index,
        # only keeping the last one.
        
        if n == 1:
            results[i] = np.transpose(df)[-1]
            var0 = np.transpose(df)[-1]
            x[i] = df[x_ind][-1]
        else:
            results[i] =np.array(df)[:,-1]
            var0 = np.array(df)[:,-1]
            x[i] = df[x_ind][-1,n_meas-1]
        
        if Tt != 0:
            x_delay = delay(t, x, Tt, value = 0)

        else:
            x_delay = x

        # control the y value by two position controller.
        y = two_position(x_delay[i], y, xlimit, ylimit)
    
    y_final[len(t)-1] = y_final[len(t)-2]
    results[len(t)-1] = results[len(t)-2]
    
    if n != 1: 
        r = np.array(results)
        results = r.transpose(1,0,2)
    else: results = np.transpose(results)
    
    return results, y_final


import numpy as np

def limit(a, a_min):

    """
    Limit the values in an array or one value.

    Args:
        a (array_like or float): Array containing elements to clip or element to clip.
        a_min (array_like or None): Minimum value in the clipped array or the clipped element.

    Returns:
        array_like: An array with the elements of a, but where values < a_min are replaced with a_min.
    """

    # limit the value with int or float type
    if isinstance(a, float) or isinstance(a, int):
        return a if a > a_min else a_min

    else:
        # check if the given minimum value larger than the maximum value in the array.
        if max(a) > a_min:
            a_max = max(a)
        else:
            a_max = a_min
        return np.clip(a, a_min, a_max)


import numpy as np
import matplotlib.pyplot as plt
import scipy as sp
from .sol_ode import sol_ode
from .get_index import get_index

def curve_fit_ode(model,
    var0,
    t,
    xdata,
    ydata,
    param=None,
    param_var0=None,
    guess0=None,
    bounds=None,
    x_ind=-1,
    n_opt = 0
    ):

    """
    Uses non-linear least squares to fit results of ODE (model) to data.

    Args:
        model (callable(y,t,...)): The function computes the derivative of y at t.
        var0 (callable or array): Initial condition of var.
        t (array): A sequence of time points for which to solve for y.
        xdata (array):  The independent variable where the data is measured.
                        Should usually be an M-length sequence or an
                        (k,M)-shaped array for functions with k predictors,
                        but can actually be any object.
        ydata (array): The dependent data, a length M array - nominally f(xdata, ...).                    
        param (array, optional): Parameters used in the ODE model function.
        param_var0 (array, optional): Parameters used in the initial condition function.
        guess0 (array, optional): Initial guess for the parameters (length N). If None,
                                  then the initial values will all be 1 (if the number of
                                  parameters for the function can be determined using
                                  introspection, otherwise a ValueError is raised).
        bounds (2-tuple of array_like, optional): Lower and upper bounds of parameters.
                                                  Defaults to [0.1*guess0, 1.5*guess0].
        x_ind (int, optional): The index of fitted results in the solved list
                               of the model. The default is the last column of
                               the list.
        n_opt (int, optional): For a model of n CSTRs, this defines which CSTR is used for the fitting. Default is the last CSTR.

    Returns:
        avg: Optimal values for the parameters so that the sum of the squared residuals
                of f(xdata, *popt) - ydata is minimized.
        cov: The estimated covariance of avg. The diagonals provide the variance of the
                parameter estimate.
    """
    
    if isinstance(param, int) | isinstance(param, float):
        param = [param]
    if isinstance(param_var0, int) | isinstance(param_var0, float):
        param_var0 = [param_var0]
    
    if isinstance(ydata, int) | isinstance(ydata, float):
        ydata = np.array([ydata] * len(xdata))
    
    if xdata[0] != 0:
        xdata.loc[-1] = 0
        xdata.index = xdata.index + 1
        xdata = xdata.sort_index()
        ydata.loc[-1] = np.math.nan
        ydata.index = ydata.index + 1
        ydata = ydata.sort_index()
                         
    if callable(var0):
        df = sol_ode(model, var0(param_var0), t, param)
    else: df = sol_ode(model, var0, t, param)
        
    if isinstance(df,list):
        C = df[x_ind]
        if isinstance(C,np.ndarray):
            if C.ndim > 1:
                C = df[x_ind][:,n_opt-1]
            else: pass
        else: pass
    elif isinstance(df,np.ndarray):
        if df.ndim > 1:
            C = df[:,n_opt-1]
        else: C = df
    else: raise('Output of the model must be list or array.')

    plt.figure('curve fit')
    plt.grid()
    plt.plot(xdata, ydata, 'o', markersize = 3)
    plt.plot(t,C,color='black')
    
    # set the default bounds
    if (bounds == None) & (guess0 != None):
        bounds = np.empty([2, len(guess0)])
        for i in range(len(guess0)):
            if guess0[i] > 0:
                bounds[0][i] = -3 * guess0[i]
                bounds[1][i] = 3 * guess0[i]
            elif guess0[i] == 0:
                bounds[0][i] = -np.inf
                bounds[1][i] = np.inf
            else:
                bounds[0][i] = 3 * guess0[i]
                bounds[1][i] = -3 * guess0[i]

    # identify the index of initial condition's parameters in input guess0
    try:
        p0, p0_converse = get_index(param_var0, guess0)
    except:
        pass

    # identify the index of ode function's parameters in input guess0
    try:
        p1, p1_converse = get_index(param, guess0)
    except:
        pass

    # create the fitted function, changing the parameter to making the ode
    # function solved results fitting the input data
    def func(xdata, *args):

        # change the parameter in the ode function
        param_new = param
        for i in range(len(p1)):
            param_new[p1_converse[i]] = args[p1[i]]

        if callable(var0):
            # change the parameter in the initial condition
            param_var0_new = param_var0
            for i in range(len(p0)):
                param_var0_new[p0_converse[i]] = args[p0[i]]
            # solve the ode function
            results = sol_ode(model, var0(param_var0_new), xdata, param_new)
        else:
            results = sol_ode(model, var0, xdata, param_new)

        # return the solved result
        if isinstance(results,list):
            ar = results[x_ind]
            if isinstance(ar,np.ndarray):
                if ar.ndim > 1:
                    return ar[:,n_opt-1][~np.isnan(ydata)]
                else: return ar[~np.isnan(ydata)]
            else: return ar[~np.isnan(ydata)]
        elif isinstance(results,np.ndarray):
            if results.ndim > 1:
                return results[:,n_opt-1][~np.isnan(ydata)]
            else: return results[~np.isnan(ydata)]
        else: raise('Output of the model must be list or array.')
        
        
    avg, cov = sp.optimize.curve_fit(func, xdata, ydata[~np.isnan(ydata)], p0=guess0, bounds=bounds)
    if callable(var0):
         df = sol_ode(model, var0(param_var0), t, param)
    else:
         df = sol_ode(model, var0, t, param)
    
        
    if isinstance(df,list):
        Cnew = df[x_ind]
        if isinstance(Cnew,np.ndarray):
            if Cnew.ndim > 1:
                Cnew = df[x_ind][:,n_opt-1]
            else: pass
        else: pass
    elif isinstance(df,np.ndarray):
        if df.ndim > 1:
            Cnew = df[:,n_opt-1]
        else: Cnew = df
    else: raise('Output of the model must be list or array.')
        
    plt.figure('curve fit')
    plt.plot(t,Cnew,color='red',linestyle='dashed')
    plt.legend(["measured data","model","fitted model"])
    
    print("Original vs fitted average values:", guess0, avg)
    std = np.sqrt(np.diag(cov))
    print("Standard deviation of estimated parameters:", std)
    
    return avg, cov


import importlib.metadata
import requests

def version():

    def get_installed_version():
        installed_version = importlib.metadata.version("sammhelper")
        return installed_version

    def check_for_updates():
        installed_version = get_installed_version()
    
        # Fetch the latest version from PyPI
        try:
            response = requests.get("https://pypi.org/pypi/sammhelper/json")
            response.raise_for_status()  # Raise an error for bad responses (e.g., 404, 500)
            latest_version = response.json()["info"]["version"]
        except requests.RequestException:
            return "Error: Unable to fetch the latest version of sammhelper from PyPI."
        
        if installed_version == latest_version:
            return f"The newest version of sammhelper ({installed_version}) is already installed."
        else:
            return (
                f"A new version of sammhelper is available ({latest_version}). "
                f"In order to update sammhelper, open your command window and enter: pip install --upgrade sammhelper"
            )


import numpy as np
import tqdm as tq
from .sol_ode import sol_ode
from .len_check import len_check
from .fill_array import fill_array

def MonteCarlo(
    model, var0, t, param=None, param_var0=None, x_ind = -1
):
    
    """
    Runs your model multiple times while varying the specified parameters.

    Args:
        model (callable(y,t,...)): The function computes the derivative of y at t.
        var0 (array): Initial condition of y.
        t (array):  A sequence of time points for which to solve for y.
        param (array, optional): Parameters used in the ODE model function.
        param_var0 (array, optional): Parameters used in the initial condition function.
        
    Returns:
        avg, std (float): Mean and standard deviation value of the all results.
        results (list): This list contains all results.
    """

    print("Start Monte Carlo simulation...")
    
    # check whether the input values for p_func, p_y0 are used for one parameter or multiple parameters
    if np.any(
        isinstance(param, (list, tuple, np.ndarray))
    ):
        var2 = np.array(param, dtype=object)  # input is a 2d array
    else:
        var2 = np.array([param], dtype=object).reshape(1, -1)  # input is a 1d array

    runs = len_check(var2)
    results = list(range(runs))

    if param_var0 != None:
        if np.any(
            isinstance(param_var0, (list, tuple, np.ndarray))
        ):
            var1 = np.array(param_var0, dtype=object)
        else:
            var1 = np.array([param_var0], dtype=object).reshape(1, -1)

        # get the numbers of runs
        runs = max(len_check(var1), len_check(var2))

        # fill the array using the runs number
        var1 = fill_array(var1, runs)

    var2 = fill_array(var2, runs)

    # different tqdm bars

    for i in tq.tqdm(range(runs)):
        var_func = [item[i] for item in var2]
        if param_var0 != None:
            var_y0 = [item[i] for item in var1]
            ydata = sol_ode(model, var0(var_y0), t, var_func)
        else:
            ydata = sol_ode(model, var0, t, var_func)
        # Workarounds for different TYPES and SIZES for ydata
        # Example 1: Exercise 08 --> TYPE list SIZE 2 --> second condition
        # Example 2: Book example 12.18 --> TYPE array SIZE (m, n) --> fourth condition
        # Example 3: Book example 12.22 --> TYPE list SIZE 1 --> first condition
        if isinstance(ydata, list) and ydata[x_ind].ndim == 1:
            results[i] = ydata[x_ind][:]
        elif isinstance(ydata, list) and ydata[x_ind].ndim > 1:
            results[i] = ydata[x_ind][:,-1]
        elif isinstance(ydata, np.ndarray) and ydata.ndim == 1:
            results[i] = ydata[:]
        else:
            results[i] = ydata[:,-1]
            
    mean = np.mean(results,0)
    stddev = np.std(results,0,ddof=1)

    return results, mean, stddev


def two_position(x, y, xlimit, ylimit):

    """
    Checks if the measured element is within its setpoints and if needed it
    adjusts the control element in order to meet this requirement. The output 
    values of the control element are in the range of its setpoints.
    
    For example: Is the oxygen concentration above xmin and below xmax?
    If no, change the aeration rate to low/OFF or high/ON in order to meet this
    requirement. 

    Args:
        x (float, optional): Measured element (for example: oxygen concentration).
        y (float, optional): Control element (for example: aeration rate).
        xmin (int, optional): Lower setpoint for measured element (for example: lower limit of oxygen concentration).
        xmax (int, optional): Upper setpoint for measured element (for example: upper limit of oxygen concentration).
        ymin (int, optional): Lower setpoint for control element (for example: low aeration rate or OFF).
        ymax (int, optional): Upper setpoint for control element (for example: high aeration rate or ON).

    Returns:
        y (float, optional): Adjusted values of the control element.
    """

    if x > xlimit[-1]:
        y = ylimit[-1]
    elif x < xlimit[0]:
        y = ylimit[0]

    return y


import matplotlib.pyplot as plt
import numpy as np
from .sol_ode import sol_ode
from .get_index import get_index
from matplotlib.ticker import MaxNLocator

def sensitivity(var, xlabel="time", model=None, var0=None, t =None, param=None, param_var0=None, x_ind = -1, linestyle='solid'):
    
    """
    Computes the absolute-relative sensitivity of the results of the model 
    to a changing parameter. Runs the model with all parameters at their 
    specified value and gets the results R1(t). Adjusts the parameter slightly 
    by adding an amount ∆ equal to 0.0001*specified value. Runs the model again, 
    calling the results R2(t). Computes the absolute-relative sensitivity S(t) 
    by the following formula: S(t) = (R2(t) - R1(t)) / ∆

    Args:
        par (float): Specified value of the parameter.
        xlabel (str): Identify the xlabel of results. 
                      There are two options: time or length. 
                      Time means results changing with time. 
                      Length means results changing along the series of reactors.
        model (callable(y,t,...)): The function computes the derivative of y at t.
        var0 (array): Initial condition of y.
        t (array): A sequence of time points for which to solve for y.
        param (array, optional): Parameters used in the model.
        param_var0 (array, optional): Parameters used in the initial condition function.
        x_ind (integer, optional): Index of the model output for the case of several outputs (e.g. X,S). Default is -1.

    Returns:
        array: Absolute-relative sensitivity of the results of the model to a changing parameter.
    """
    
    # Set up the line style based on plot count
    line_styles = ['solid', 'dashed', 'dashdot', 'dotted']
    
    if isinstance(param, int) | isinstance(param, float):
        param = [param]
    if isinstance(param_var0, int) | isinstance(param_var0, float):
        param_var0 = [param_var0]
    
    if param_var0 != None:

        # solve the ode function using the initial input parameter
        C = sol_ode(model, var0(param_var0), t, param)

        # times the parameter with 1.0001
        param_var0_new = param_var0
        param_new = param
        if var in param_var0:
            index = get_index(var, param_var0)
            param_var0_new[index[0]] = 1.0001 * var
        if var in param:
            index = get_index(var, param)
            param_new[index[0]] = 1.0001 * var

        # solve the ode function using the changed parameter
        C_sens = sol_ode(model, var0(param_var0_new), t, param_new)
    else:
        # solve the ode function using the initial input parameter
        C = sol_ode(model, var0, t, param)

        # times the parameter with 1.0001
        param_new = param

        if var in param:
            index = get_index(var, param)
            param_new[index[0]] = 1.0001 * var

        # solve the ode function using the changed parameter
        C_sens = sol_ode(model, var0, t, param_new)
    # calculate the sensitivity of the parameter along the time or length
    if xlabel == "time":
        if var == 0:
            try:
                result = var*(C_sens - C) / (0.0001)
            except:
                result = var*(C_sens[x_ind]- C[x_ind])/(0.0001)
        else:
            try:
                result = var*(C_sens - C) / (0.0001 * var)
            except:
                result = var*(C_sens[x_ind]- C[x_ind])/(0.0001 * var)        
        # plot the results
        plt.figure('sensitivity')
        plt.grid()
        try:
            plt.plot(t,result[:,-1], linestyle = linestyle)
        except:
            plt.plot(t,result, linestyle = linestyle)
        
        return result
    
    elif xlabel == "length":
        if var == 0: 
            result = var * (np.array([arr[-1] for arr in C_sens]) - np.array([arr[-1] for arr in C])) / 0.0001
        else:
            result = var * (np.array([arr[-1] for arr in C_sens]) - np.array([arr[-1] for arr in C])) / (0.0001 * var)
        
        # plot the results
        plt.figure('sensitivity')
        plt.plot(range(1,len(result)+1), result, linestyle = linestyle)
        plt.gca().xaxis.set_major_locator(MaxNLocator(integer=True))
        plt.xlim(xmin=1, xmax=len(result)+1)
          
        return result


from datetime import datetime

def greet():
    
    """
    Greets user based on time of the day.
    
    Returns: 
        greetings (str)
    """ 
    
    time = datetime.now().hour
    if time < 12:
        return "Good Morning!"
    elif time < 15:
        return "Good Afternoon!"
    elif time < 18:
        return "Good Evening!"
    else:
        return "Good Night!"


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import scipy as sp
import odeintw as ow
import tqdm as tq

def sol_ode(model, var0, t, param):

    """
    Solves a system of ordinary differential equations (ODEs).
    
    Args:
        model (callable(var,t,param)): The function computes the derivative of y at t.
        var0 (array): Initial condition of var.
        t (array): A sequence of time points at which var is calculated.
        param (array): Model parameters.

    Returns:
        ar (array): Array containing the value of var for each desired time in t, with the initial value var0 in the first row.
    """
    
    dt = t[1]-t[0]
    if isinstance(param,list):
        if (len(param) == 1):
            param = param[0]
        else: pass
    else: pass

    if isinstance(var0,list):
        var0 = np.array(var0)
    else: pass
    
    results = ow.odeintw(model, y0=var0, t=t, args=(param,),hmax = dt)
    
    try:
        out = model(var0, t, param) #this is a workaround, so in the case of 1 variable and n CSTRs the result does not get flipped
    except ValueError:
        out = (1,1)
    
    if isinstance(out, tuple):
        m = int(np.shape(var0)[0])
        ar = list(range(m))
        for i in range(m):
            ar[i] = results[:,i]
    else:
        ar = results
    return ar


import numpy as np

def fill_array(a, l):

    """
    Fill each row in the two-dimensional (2d) array so that the rows of the array have the same length l.

    Args:
        a (2d-array): Input 2d array.
        l (int): Row Length to be filled.

    Returns:
        a (2d-array): The filled 2d array of equal length rows.
    """

    for i in range(len(a)):
        if not (isinstance(a[i], (list, np.ndarray))):
            a[i] =l * [a[i]]

    return a


import numpy as np
import scipy as sp
import pandas as pd

def fft(data):

    """
    Computes the 1-D discrete Fourier Transformation. 
    
    Args:
        data (dataframe or list): The dataframe or list with two columns, first: time, second: value.
    
    Returns:
        frq (array): x-axis (= frequency) of the descrete frequency spectrum. 
        amp (array): y-axis (= amplitude) of the descrete frequency spectrum. 
    """

    # acess data frame or list. Raise error if neither is given.
    if isinstance(data,pd.DataFrame):
        x = data.iloc[:, 0]
        y = data.iloc[:, 1]
    elif isinstance(data, list):
        x = data[0]
        y = data[1]
    else: raise Exception('Possible input: pd.DataFrame or list')
    
    # sampling interval
    si = np.median(np.diff(x))
    
    # sampling frequency
    sf = 1/si
    
    # frequencies (x-axis)
    N = len(y)
    k = np.arange(int(N/2))
    p  = N/sf
    frq = k/p
    
    # amplitude (y-axis)
    amp = sp.fft.fft(np.array(y))
    
    # normalize amplitude
    amp = abs(amp/len(y)*2)
    amp[0] = amp[0]/2
    
    # exclude sampling frequency from amplitude
    amp = amp[range(int(len(y)/2))]
    
    # output
    return frq, amp


import numpy as np

def delay(time, var, delay, value = 0):

    """
    This function returns a sequence of values shifted by the delay value.

    Args:
        time (array): A sequence of time points.
        var (array): A sequence of values.
        delay (float): Specified delay time.
        value (float, optional): Specified value for var for the time before the delay time. 
        

    Returns:
        var_delayed (array): Returns a sequence of values delayed by the delay time with the same length as the time vector.
    """
    
    DT = time[1]-time[0]
    i_delay = int(round(delay/DT))
    if np.ndim(var)!= 1: var = var[:,-1]
    else: pass
    
    var_delayed = np.zeros(len(var))
    var_delayed[0:i_delay] = value
    var_delayed[i_delay:] = var[0:-i_delay]
    
    return var_delayed


import numpy as np
    
def nextprime(n):
    
    """
    Finds the next prime number.
    
    Args:
        n (int): Any number.
    
    Returns:
        p (int): Next prime number.
    """ 
    
    p=n+1
    for i in range(2,p):
        if(np.mod(p,i)==0):
            p=p+1
    else:
        return p


import pandas as pd

def data_import(file,columnnames):

    """
    Imports a file and renames the columns.

    Args:
        file (str): The location of the file.

    Returns:
        data (Dataframe): In case of SAMM the table mostly contains time and concentration values.
    """

    data = pd.read_csv(file, sep="\t|,|;", engine='python')

    # set the column name
    data.columns = columnnames
    print("The overview of the dataset:\n", data)
    return data


import numpy as np

def len_check(a):

    """
    Get the maximum row length in a two-dimensional array.

    Args:
        a (2d-array): Input 2d array.

    Returns:
        max_len (int): Maximum row length in 2d-array a.
    """

    max_len = 0
    for item in a:
        if isinstance(item, (list, np.ndarray)):
            if len(item) > max_len:
                max_len = len(item)

    return max_len


from .greet import greet
from .nextprime import nextprime
from .sol_ode import sol_ode
from .data_import import data_import
from .curve_fit_ode import curve_fit_ode
from .curve_fit import curve_fit
from .delay import delay
from .sensitivity import sensitivity
from .limit import limit
from .MonteCarlo import MonteCarlo
from .fft import fft
from .two_pos_controller import two_pos_controller
from .PID_controller import PID_controller
from .version import version